#include <sys/types.h>
#include <fcntl.h>
#include <stddef.h>
#include <osreldate.h>
#if __FreeBSD_version < 500000
# include <libusb.h>
#else
# include <usbhid.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <dev/usb/usb.h>
#include <dev/usb/usbhid.h>

#include "usb_ctrl.h"
#include "usb_ctrl_def.h"

#include "CommandPost.h"

#include "../dvview.h"

#include <stdio.h>

struct item {
    int32_t		usage;
#ifdef DEBUG
    char		*usage_name;
#endif
    struct hid_item	*h;
    int32_t		data;
#ifdef DEBUG
    int32_t		diff;
#endif
    struct ctrl_bind	*bind;
};


static int num_item;
static int dlen = 0;
static u_char *dbuf = NULL;
static struct hid_item *hids;
static struct item *itemlist;

int
controller_init(const char *devname)
{
    int fd, cnt;
    struct hid_data *d;
    struct hid_item h, *n;
    u_int32_t colls[100];
    report_desc_t rd;
    struct ctrl_bind *bind;
    char buf1[30], buf2[30], *str, *ptr;

    int valid;

    hid_init(NULL);

    if ((fd = open(devname, O_RDONLY)) < 0) {
	warn(devname);
	return fd;
    }

    if ((rd = hid_get_report_desc(fd)) == 0) {
	warn("USB_GET_REPORT_DESC");
	return -1;
    }
    hids = 0;
    cnt = 0;
    num_item = 0;
#if __FreeBSD_version < 500000
    d = hid_start_parse(rd, 1<<hid_input);
#else
    d = hid_start_parse(rd, 1<<hid_input, 0);
#endif
    while (hid_get_item(d, &h)) {
	if (h.kind == hid_collection) {
	    colls[++cnt] = h.usage;
	} else if (h.kind == hid_endcollection) {
	    --cnt;
	}
	if (h.kind != hid_input || (h.flags & HIO_CONST)) {
	    continue;
	}
	h.next = hids;
	h.collection = colls[cnt];
	hids = malloc(sizeof *hids);
	*hids = h;
	++num_item;
    }
    hid_end_parse(d);
    dlen = hid_report_size(rd, 0, hid_input);
    dbuf = malloc(dlen);

    printf("HID: %d item found\n", num_item);
    itemlist = malloc(sizeof(struct item) * num_item);
    cnt = 0;
    valid = 0;
    for (n = hids; n; n = n->next) {
	strcpy(buf1, hid_usage_in_page(n->usage));
	str = buf1;
	if ((ptr = strchr(buf1, ' ')) != NULL) {
	    do {
		*ptr++ = '\0';
	    } while (*ptr == ' ');
	    sprintf(buf2, buf1, atoi(ptr));
	    str = buf2;
	}
	itemlist[cnt].usage = n->usage;
#ifdef DEBUG
	itemlist[cnt].usage_name = strdup(str);
#endif
	itemlist[cnt].h = n;
	itemlist[cnt].data = 0;
#ifdef DEBUG
	itemlist[cnt].diff = 0;
#endif
	itemlist[cnt].bind = NULL;
	for (bind = CommandPost_def.bind; bind->usage_name != NULL; bind++) {
	    if (!strcmp(bind->usage_name, str)) {
		itemlist[cnt].bind = bind;
		++valid;
		break;
	    }
	}
	printf("%d: %06x, %s (%s)\n",
	       cnt, itemlist[cnt].usage, str,
	       itemlist[cnt].bind == NULL ? "---" : bind->name);
	++cnt;
    }
    printf("HID: %d actions are activated\n", valid);

    return fd;
}


int
controller_get_event(int fd)
{
    int data, diff, action;
    int cnt, cnt2, tmp;
    struct hid_item *n;
    trans_result_t trans_result;

    if ((tmp = read(fd, dbuf, dlen)) != dlen) {
	err(1, "bad read %d != %d", tmp, dlen);
    }
    action = ACT_NONE;
    for (cnt = 0; action == ACT_NONE && cnt < num_item; cnt++) {
	n = itemlist[cnt].h;
	if (itemlist[cnt].bind == NULL) {
	    continue;
	}
#ifdef DEBUG
	fputs(itemlist[cnt].usage_name, stdout);
	printf(" (%04x)", n->usage);
#endif
	for (cnt2 = 0; action == ACT_NONE && cnt2 < n->report_count; cnt2++) {
	    if (n->logical_minimum < 0) {
		data = (int)hid_get_data(dbuf, n);
	    } else {
		data = (u_int)hid_get_data(dbuf, n);
	    }
	    if (itemlist[cnt].bind->func != NULL) {
		trans_result = itemlist[cnt].bind->func(data,
							itemlist[cnt].data);
		data = trans_result.data;
#ifdef DEBUG
		diff = trans_result.diff;
#endif
		action = trans_result.action;
#ifdef DEBUG
		if (action != ACT_NONE) {
		    printf(" (changed)<%04x>", action);
		}
#endif
	    } else {
		diff = data - itemlist[cnt].data;
		if (diff != 0) {
		    if (0 < diff) {
			action = itemlist[cnt].bind->action_make;
		    } else {
			action = itemlist[cnt].bind->action_break;
		    }
#ifdef DEBUG
		    printf(" (changed)[%d]<%04x>", diff, action);
#endif
		    break;
		}
	    }
#ifdef DEBUG
	    printf(" %d", data);
#endif
	    itemlist[cnt].data = data;
#ifdef DEBUG
	    itemlist[cnt].diff = diff;
#endif
	}
#ifdef DEBUG
	putchar('\n');
#endif
    }

    return action;
}
