#include <sys/types.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <curses.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "dvview.h"
#include "ctrl/usb_ctrl.h"


static const char rcsid[] = "$Id: dvview.c,v 1.12 2004/10/04 03:37:51 a-gota Exp $";

#define	max(a, b)	(a < b ? b : a)

#define	PLAYDV_PATH	"/usr/local/bin/playdv"
#define	FWCONTROL_CMD	"/usr/sbin/fwcontrol -S /dev/stdin"
#define	DIF_BLK_SIZE		    80
#define	DIF_NUM_BLKS_IN_SEQ	   150
#define	DIF_SEQ_SIZE		(DIF_BLK_SIZE * DIF_NUM_BLKS_IN_SEQ)
#define	NUM_DIF_SEQ_525		    10
#define	NUM_DIF_SEQ_625		    12
#define	MAX_FRAME_SIZE	(DIF_SEQ_SIZE * max(NUM_DIF_SEQ_525, NUM_DIF_SEQ_625))

#define	UHID_DEVNAME	"/dev/uhid0"
#define	UHID_DEVTYPE	"CommandPost"


#define	NUM_PGM		20

typedef struct {
    int		bank_no;
    long	start_frame, end_frame;
} pgm_item_t;

pgm_item_t	pgm[NUM_PGM];
int		num_of_pgm, cur_pgm;


#define	NUM_BANK	 8

struct stream_bank {
    int			no;
    char		*filename;
    FILE		*fp;
    int			system_is_625;
    long		frame_size, num_frame;
    int			last_pos;
    struct timeval	*timeout;
} bank[NUM_BANK], *cur_bank = NULL;

struct timeval	timeout_0, timeout_50, timeout_60;
u_char *frame = NULL;
pid_t playdv_pid = 0, fwcontrol_pid = 0;
WINDOW *win_action = NULL, *win_frame_info = NULL, *win_message = NULL;
WINDOW *win_status = NULL, *win_streams = NULL, *win_dialog = NULL;
WINDOW *win_mode1 = NULL, *win_mode2 = NULL;
WINDOW *win_pgm = NULL, *win_pgm_items = NULL;
int pgm_win_lines;
FILE *fp_stdin;
struct termios t_orig, t_raw;
char *recdate, *rectime, *source;


int dumb_get_char(FILE *tty_fp) {
    return fgetc(tty_fp);
}


void dumb_putn(const char *msg, int len) {
    fwrite(msg, len, 1, stdout);
}


void dumb_print(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
    putchar('\n');
}


void dumb_print_nonl(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
}


void dumb_print_term(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
    putchar('\n');
}


void dumb_put_frame_no(long frame_no) {
    printf("#%05ld", frame_no);
}


void dumb_put_timecode(int hh, int mm, int ss, int fr, int sys_625, int df) {
    printf(": %02d:%02d:%02d.%02d", hh, mm, ss, fr);
    if (!sys_625) {
	fputs(df ? " DF" : " NDF", stdout);
    }
}


void dumb_put_datacode(const char *recdate, const char *rectime,
		       const char *source) {
    fputs("    ", stdout);
    fputs(recdate, stdout);
    putchar(' ');
    fputs(rectime, stdout);
    fputs("  ", stdout);
    fputs(source, stdout);
}


void dumb_finish_frame_info(void) {
    putchar('\r');
    fflush(stdout);
}


void dumb_clear_frame_info(void) {
    putchar('\n');
}


void dumb_warn(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vwarn(fmt, ap);
    va_end(ap);
}


void dumb_clear_status(void) {
    putchar('\n');
}


void dumb_put_stream_info(struct stream_bank *bank) {
    if (bank->fp != NULL) {
	printf("DV stream #%d is opened; %s, ", bank->no + 1, bank->filename);
	if (bank->system_is_625 == 0) {
	    printf("525/60 system");
	} else {
	    printf("625/50 system");
	}
	printf(", %ld frame(s)\n", bank->num_frame);
    }
}


void dumb_put_selected_stream(struct stream_bank *bank, int selected) {
    return;
}


void dumb_get_string(char *buf, int len,
		       const char *title, const char *msg, ...) {
    va_list ap;
    char *ptr;

    tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_orig);
    printf("\n%s\n", title);
    va_start(ap, msg);
    vfprintf(stdout, msg, ap);
    va_end(ap);
    fflush(stdout);
    fgets(buf, len, fp_stdin);
    tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_raw);

    if ((ptr = strchr(buf, '\n')) != NULL) {
	*ptr = '\0';
    }
}


void dumb_disp_pgm(void) {
    int cnt;

    printf("\n%d program(s) valid\n", num_of_pgm);
    printf("Pgm #\tStr #\t   IN\t   OUT\n");
    for (cnt = 0; cnt < num_of_pgm; cnt++) {
	printf("% 4d\t% 4d\t", cnt + 1, pgm[cnt].bank_no + 1);
	if (pgm[cnt].start_frame < 0) {
	    printf("   ----");
	} else {
	    printf("% 7ld", pgm[cnt].start_frame);
	}
	if (pgm[cnt].end_frame < 0) {
	    printf("\t   ----\n");
	} else {
	    printf("\t% 7ld\n", pgm[cnt].end_frame);
	}
    }
}


void dumb_clean_up(void) {
    printf("\nClean up...\n");
}


int curses_get_char(FILE *tty_fp /* unused */) {
    return getch();
}


void curses_put_action_norefresh(const char *fmt, ...) {
    va_list ap;

    wmove(win_action, 0, 0);
    va_start(ap, fmt);
    vwprintw(win_action, (char *)fmt, ap);
    va_end(ap);
}


void curses_put_action(const char *fmt, ...) {
    va_list ap;

    wmove(win_action, 0, 0);
    va_start(ap, fmt);
    vwprintw(win_action, (char *)fmt, ap);
    va_end(ap);
    wclrtoeol(win_action);
    wrefresh(win_action);
}


void curses_add_action_norefresh(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vwprintw(win_action, (char *)fmt, ap);
    va_end(ap);
}


void curses_add_action(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vwprintw(win_action, (char *)fmt, ap);
    va_end(ap);
    wclrtoeol(win_action);
    wrefresh(win_action);
}


void curses_put_frame_no(long frame_no) {
    mvwprintw(win_frame_info, 0, 0, "#%05ld", frame_no);
    wclrtoeol(win_frame_info);
}


void curses_put_timecode(int hh, int mm, int ss, int fr, int sys_625, int df) {
    mvwprintw(win_frame_info, 1, 0, "%02d:%02d:%02d.%02d", hh, mm, ss, fr);
    if (!sys_625) {
	waddstr(win_frame_info, df ? " DF" : " NDF");
    }
    wclrtoeol(win_frame_info);
}


void curses_put_datacode(const char *recdate, const char *rectime,
			 const char *source) {
    mvwaddstr(win_frame_info, 0, 20, recdate);
    mvwaddstr(win_frame_info, 0, 30, rectime);
    mvwaddstr(win_frame_info, 1, 20, source);
    wclrtoeol(win_frame_info);
}


void curses_finish_frame_info(void) {
    wrefresh(win_frame_info);
}


void curses_clear_frame_info(void) {
    wclear(win_frame_info);
    wrefresh(win_frame_info);
}


void curses_put_status_norefresh(const char *fmt, ...) {
    va_list ap;

    wmove(win_status, 0, 0);
    va_start(ap, fmt);
    vwprintw(win_status, (char *)fmt, ap);
    va_end(ap);
}


void curses_put_status(const char *fmt, ...) {
    va_list ap;

    wmove(win_status, 0, 0);
    va_start(ap, fmt);
    vwprintw(win_status, (char *)fmt, ap);
    va_end(ap);
    wclrtoeol(win_status);
    wrefresh(win_status);
}


void curses_add_status_norefresh(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vwprintw(win_status, (char *)fmt, ap);
    va_end(ap);
}


void curses_add_status(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vwprintw(win_status, (char *)fmt, ap);
    va_end(ap);
    wclrtoeol(win_status);
    wrefresh(win_status);
}


void curses_put_warn(const char *fmt, ...) {
    va_list ap;

    wmove(win_status, 0, 0);
    va_start(ap, fmt);
    vwprintw(win_status, (char *)fmt, ap);
    va_end(ap);
    waddstr(win_status, ": ");
    waddstr(win_status, strerror(errno));
    wclrtoeol(win_status);
    wrefresh(win_status);
}


void curses_clear_status(void) {
    wclear(win_status);
    wrefresh(win_status);
}


void curses_put_message(const char *msg, int len) {
    waddnstr(win_message, msg, len);
    wrefresh(win_message);
}


void curses_show_stream_info(struct stream_bank *bank) {
    if (bank->fp != NULL) {
	mvwprintw(win_status, 0, 0, "DV stream #%d is opened", bank->no + 1);
	wclrtoeol(win_status);
	wrefresh(win_status);
    }
    mvwprintw(win_streams, bank->no + 3, 4, "%c: ", bank->no + 'A');
    if (bank->fp != NULL) {
	waddstr(win_streams, bank->filename);
	waddstr(win_streams, ", ");
	if (bank->system_is_625 == 0) {
	    waddstr(win_streams, "525/60");
	} else {
	    waddstr(win_streams, "625/50");
	}
	wprintw(win_streams, " system, %ld frame(s)", bank->num_frame);
    }
    wclrtoeol(win_streams);
    wrefresh(win_streams);
}


void curses_mark_stream_selected(struct stream_bank *bank, int selected) {
    mvwaddch(win_streams, bank->no + 3, 1, selected ? '>' : ' ');
    wrefresh(win_streams);
}


void curses_put_mode1(const char *mode) {
    waddnstr(win_mode1, mode, 3);
    wrefresh(win_mode1);
}


void curses_clear_mode1(void) {
    werase(win_mode1);
    wrefresh(win_mode1);
}


void curses_put_mode2(const char *mode) {
    waddnstr(win_mode2, mode, 3);
    wrefresh(win_mode2);
}


void curses_clear_mode2(void) {
    werase(win_mode2);
    wrefresh(win_mode2);
}


void curses_dialog(char *buf, int len,
			    const char *title, const char *msg, ...) {
    WINDOW *dialog;
    va_list ap;
    int ret;
    char *ptr;

    if ((win_dialog = newwin(4, COLS - 10, (LINES - 4) / 2, 10 / 2)) == NULL) {
	err(3, "newwin(dialog)");
    }
    box(win_dialog, 0, 0);
    mvwaddch(win_dialog, 0, 2, ' ');
    waddstr(win_dialog, title);
    waddch(win_dialog, ' ');
    va_start(ap, msg);
    wmove(win_dialog, 1, 2);
    vwprintw(win_dialog, (char *)msg, ap);
    va_end(ap);
    wrefresh(win_dialog);
    if ((dialog = derwin(win_dialog, 2, COLS - 10 - 22 - 2, 1, 22)) == NULL) {
	err(4, "derwin(dialog)");
    }
    scrollok(dialog, TRUE);
    touchwin(win_dialog);
    wrefresh(dialog);

    nocbreak();
    echo();
    ret = mvwgetnstr(dialog, 0, 0, buf, len);
    noecho();
    cbreak();

    delwin(dialog);
    delwin(win_dialog);
    win_dialog = NULL;

    touchwin(stdscr);
    refresh();
    if (win_pgm != NULL) {
	touchwin(win_pgm);
	wrefresh(win_pgm);
    }

    if ((ptr = strchr(buf, '\n')) != NULL) {
	*ptr = '\0';
    }
}


void curses_disp_pgm(void) {
    void curses_refresh_pgm(void);
    void curses_hide_pgm(void);

    if (win_pgm == NULL) {
	pgm_win_lines = LINES - 7;
	if ((win_pgm = newwin(pgm_win_lines + 4, 26, 2, COLS - 26)) == NULL) {
	    pgm_win_lines = 0;
	    err(3, "newwin(pgm)");
	}
	box(win_pgm, 0, 0);
	mvwaddstr(win_pgm, 0, 2, "Program List");
	mvwprintw(win_pgm, 1, 1, "Pgm  Bnk     Frame No.");
	mvwprintw(win_pgm, 2, 1, "No.  No.    IN      OUT");
/*	mvwprintw(win_pgm, n, 1, "nnn: nnn  nnnnnn  nnnnnn"); */
	leaveok(win_pgm, TRUE);
	wrefresh(win_pgm);
	if ((win_pgm_items = derwin(win_pgm, pgm_win_lines, 24, 3, 1)) == NULL) {
	    err(4, "derwin(pgm_items)");
	}
	leaveok(win_pgm_items, TRUE);
	curses_refresh_pgm();
    } else {
	curses_hide_pgm();
    }
}


void curses_refresh_pgm(void) {
    int cnt, top;

    if (win_pgm == NULL) {
	curses_disp_pgm();
    } else {
	top = cur_pgm - pgm_win_lines / 2;
	if (num_of_pgm <= top + pgm_win_lines) {
	    top = num_of_pgm - pgm_win_lines;
	}
	if (top < 0) {
	    top = 0;
	}
	for (cnt = 0; cnt < pgm_win_lines; cnt++) {
	    if (num_of_pgm <= top + cnt) {
		break;
	    }
/*	    mvwprintw(win_pgm, n, 1, "nnn: nnn  nnnnnn  nnnnnn"); */
	    mvwprintw(win_pgm_items, cnt, 0, "% 3d%c % 3d",
		      top + cnt + 1, (top + cnt == cur_pgm ? '>' : ':'),
		      pgm[top + cnt].bank_no + 1);
	    if (pgm[top + cnt].start_frame == EOF) {
		wprintw(win_pgm_items, "    ----");
	    } else {
		wprintw(win_pgm_items, "  %06d", pgm[top + cnt].start_frame);
	    }
	    if (pgm[top + cnt].end_frame == EOF) {
		wprintw(win_pgm_items, "    ----");
	    } else {
		wprintw(win_pgm_items, "  %06d", pgm[top + cnt].end_frame);
	    }
	}
	if (cnt < pgm_win_lines) {
	    wclrtoeol(win_pgm_items);
	}
	touchwin(win_pgm);
	wrefresh(win_pgm_items);
    }
}


void curses_hide_pgm(void) {
    if (win_pgm != NULL) {
	delwin(win_pgm_items);
	delwin(win_pgm);
	win_pgm_items = NULL;
	win_pgm = NULL;
	pgm_win_lines = 0;
	touchwin(stdscr);
	refresh();
    }
}


void curses_cursor_off(void) {
    move(LINES - 1, COLS - 1);
    curs_set(0);
    refresh();
}

void curses_clean_up(void) {
    curses_hide_pgm();
    wclear(win_action);
    wrefresh(win_action);
    wclear(win_frame_info);
    wrefresh(win_frame_info);
    wclear(win_status);
    wrefresh(win_status);
    wclear(win_streams);
    wrefresh(win_streams);
    endwin();
    printf("Clean up...\n");
}


void do_nothing(void) {
    return;
}


struct cons_handler {
    int	 (*get_char)(FILE *tty_fp);
    void (*put_action)(const char *fmt, ...);
    void (*put_action_nonl)(const char *fmt, ...);
    void (*put_action_cont)(const char *fmt, ...);
    void (*put_action_term)(const char *fmt, ...);
    void (*put_frame_no)(long frame_no);
    void (*disp_timecode)(int hh, int mm, int ss, int fr, int sys_625, int df);
    void (*disp_datacode)(const char *recdate, const char *rectime,
			  const char *source);
    void (*finish_frame_info)(void);
    void (*clear_frame_info)(void);
    void (*put_status)(const char *fmt, ...);
    void (*put_status_nonl)(const char *fmt, ...);
    void (*put_status_cont)(const char *fmt, ...);
    void (*put_status_term)(const char *fmt, ...);
    void (*warn)(const char *fmt, ...);
    void (*clear_status)(void);
    void (*put_message)(const char *msg, int len);
    void (*show_stream_info)(struct stream_bank *bank);
    void (*mark_stream_selected)(struct stream_bank *bank, int selected);
    void (*put_mode1)(const char *mode);
    void (*clear_mode1)(void);
    void (*put_mode2)(const char *mode);
    void (*clear_mode2)(void);
    void (*filename_dialog)(char *buf, int len, const char *title,
			    const char *msg, ...);
    void (*dialog)(char *buf, int len, const char *title,
			    const char *msg, ...);
    void (*pgm_disp)(void);
    void (*pgm_refresh)(void);
    void (*pgm_hide)(void);
    void (*cursor_off)(void);
    void (*clean_up)(void);
} *cons, dumb_handler = {
    dumb_get_char,
    dumb_print,
    dumb_print_nonl,
    dumb_print_nonl,
    dumb_print_term,
    dumb_put_frame_no,
    dumb_put_timecode,
    dumb_put_datacode,
    dumb_finish_frame_info,
    dumb_clear_frame_info,
    dumb_print,
    dumb_print_nonl,
    dumb_print_nonl,
    dumb_print_term,
    dumb_warn,
    dumb_clear_status,
    dumb_putn,
    dumb_put_stream_info,
    dumb_put_selected_stream,
    (void (*)(const char *))dumb_print,
    do_nothing,
    (void (*)(const char *))dumb_print,
    do_nothing,
    dumb_get_string,
    dumb_get_string,
    dumb_disp_pgm,
    dumb_disp_pgm,
    do_nothing,
    do_nothing,
    dumb_clean_up
}, curses_handler = {
    curses_get_char,
    curses_put_action,
    curses_put_action_norefresh,
    curses_add_action_norefresh,
    curses_add_action,
    curses_put_frame_no,
    curses_put_timecode,
    curses_put_datacode,
    curses_finish_frame_info,
    curses_clear_frame_info,
    curses_put_status,
    curses_put_status_norefresh,
    curses_add_status_norefresh,
    curses_add_status,
    curses_put_warn,
    curses_clear_status,
    curses_put_message,
    curses_show_stream_info,
    curses_mark_stream_selected,
    curses_put_mode1,
    curses_clear_mode1,
    curses_put_mode2,
    curses_clear_mode2,
    curses_dialog,
    curses_dialog,
    curses_disp_pgm,
    curses_refresh_pgm,
    curses_hide_pgm,
    curses_cursor_off,
    curses_clean_up
};


void
clean_up(void)
{
    int tmp;

    cons->clean_up();
    for (tmp = 0; tmp < NUM_BANK; tmp++) {
	if (bank[tmp].filename != NULL) {
	    free(bank[tmp].filename);
	}
	if (bank[tmp].fp != NULL) {
	    fclose(bank[tmp].fp);
	}
    }
    if (frame != NULL) {
	free(frame);
    }
    if (0 < playdv_pid) {
	kill(playdv_pid, SIGTERM);
	waitpid(playdv_pid, &tmp, WNOHANG);
    }
    if (0 < fwcontrol_pid) {
	kill(fwcontrol_pid, SIGTERM);
	waitpid(fwcontrol_pid, &tmp, WNOHANG);
    }
}

#define MSGBUFLEN	65536
int msgpipefd[2] = { -1, -1 };
int msglen = 0;
char msgbuf[MSGBUFLEN];

#ifdef MSG_BY_SIGIO
void msg_arrive(int sig)
{
    int len;

    if (sig != SIGIO) {
	return;
    }

    if (0 < msgpipefd[0]) {
	fcntl(msgpipefd[0], F_SETFL, O_NONBLOCK);
/*
	while (msglen < MSGBUFLEN && 0 < (len = read(msgpipefd[0], &msgbuf[msglen], MSGBUFLEN - msglen))) {
*/
	while (0 < (len = read(msgpipefd[0], msgbuf, MSGBUFLEN))) {
	    cons->put_message(msgbuf, len);
/*
	    msglen += len;
*/
	}
    }
/*
    if (msglen < MSGBUFLEN) {
	fcntl(msgpipefd[0], F_SETFL, O_NONBLOCK | O_ASYNC);
    }
    signal(SIGIO, msg_arrive);
*/
    fcntl(msgpipefd[0], F_SETFL, O_ASYNC);
}
#endif /* MSG_BY_SIGIO */


int
main(int argc, char *argv[])
{
    char buf[FILENAME_MAX], *ptr;
    int num_opened, bankno;
    FILE *fp;
    FILE *fi, *fo;
    int controller_fd, num_fd;
    int command, play_f, eof_f, exit_f = 0, pipefd[2], verbose_f = 0;
    int outpipe[2];
    int repeat_f, pgmmode_f;
    int use_curses_f = 0;
    int param, skip, div, wait_frame;
    int dir;
    long frame_size, num_frame, frame_no = 0;
    long displayed_no = 0;
    long start_frame, end_frame;
    fd_set readfds;
    int c, nodisp_f = 0, error_f = 0;
    int cnt;
    char *playdv_path = PLAYDV_PATH;
    char *fwcontrol_cmd = FWCONTROL_CMD;
    char *uhid_path = UHID_DEVNAME, *uhid_type = UHID_DEVTYPE;
    char *listfile = NULL;
    int tmp1, tmp2;
    long ltmp1, ltmp2;

    FILE *open_stream(struct stream_bank *bank, const char *filename);
    int load_list(const char *filename);
    void save_list(const char *filename);
    int get_key_event(FILE *fp);
    void parse_subcode(u_char *frame, int system_is_625);
    void usage(char *argv0);

    while ((c = getopt(argc, argv, "c:f:hl:np:vw")) != EOF) {
	switch (c) {
	  case 'c':
	    if ((ptr = strchr(optarg, ':')) != NULL) {
		*ptr = '\0';
		uhid_type = ptr + 1;
	    }
	    if (optarg != ptr) {
		uhid_path = optarg;
	    }
	    break;
	  case 'f':
	    fwcontrol_cmd = optarg;
	    break;
	  case 'h':
	    ++error_f;
	    break;
	  case 'l':
	    listfile = optarg;
	    break;
	  case 'n':
	    ++nodisp_f;
	    break;
	  case 'p':
	    playdv_path = optarg;
	    break;
	  case 'v':
	    ++verbose_f;
	    break;
	  case 'w':
	    ++use_curses_f;
	    break;
	  default:
	    ++error_f;
	    break;
	}
    }

    if (error_f) {
	usage(argv[0]);
    }

    argc -= optind;
    argv += optind;

    pipe(pipefd);
    pipe(msgpipefd);
    if ((playdv_pid = fork()) == 0) {
	/* child */
	dup2(pipefd[0], 0);
	close(pipefd[0]);
	dup2(msgpipefd[1], 1);
	dup2(msgpipefd[1], 2);
	close(msgpipefd[1]);
	execl(playdv_path, playdv_path, NULL);
	return 1;
    }
    /* parent */
    close(pipefd[0]);

    if ((frame = malloc(MAX_FRAME_SIZE)) == NULL) {
	kill(playdv_pid, SIGTERM);
	waitpid(playdv_pid, &command, WNOHANG);
	err(1, "malloc for frame memory");
    }

    controller_fd = controller_init(uhid_path);

#ifdef MSG_BY_SIGIO
    signal(SIGIO, msg_arrive);
    fcntl(msgpipefd[0], F_SETOWN, getpid());
    fcntl(msgpipefd[0], F_SETFL, O_NONBLOCK | O_ASYNC);
#endif

    fp_stdin = fopen("/dev/tty", "r");
    if (use_curses_f) {
	initscr();
	newterm(getenv("TERM"), stdout, fp_stdin);
	savetty();
	noecho();
	cbreak();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	if ((win_action = subwin(stdscr, 1, 20, 0, 0)) == NULL) {
	    err(2, "newwin(action)");
	}
	leaveok(win_action, TRUE);
	touchwin(stdscr);
	wrefresh(win_action);
	if ((win_frame_info = subwin(stdscr, 2, 42, 2, 0)) == NULL) {
	    err(2, "newwin(frame_info)");
	}
	leaveok(win_frame_info, TRUE);
	touchwin(stdscr);
	wrefresh(win_frame_info);
	if ((win_mode1 = subwin(stdscr, 1, 3, 0, COLS - 3)) == NULL) {
	    err(2, "newwin(mode1)");
	}
	leaveok(win_mode1, TRUE);
	touchwin(stdscr);
	wrefresh(win_mode1);
	if ((win_mode2 = subwin(stdscr, 1, 3, 1, COLS - 3)) == NULL) {
	    err(2, "newwin(mode2)");
	}
	leaveok(win_mode2, TRUE);
	touchwin(stdscr);
	wrefresh(win_mode2);
	if ((win_message = subwin(stdscr,
				  LINES - (NUM_BANK + 4 + 5), COLS,
				  5, 0)) == NULL) {
	    err(2, "newwin(message)");
	}
	scrollok(win_message, TRUE);
	wsetscrreg(win_message, 0, LINES - (NUM_BANK + 4 + 6) - 1);
	leaveok(win_message, TRUE);
	touchwin(stdscr);
	wrefresh(win_message);
	if ((win_streams = subwin(stdscr,
				  NUM_BANK + 4, COLS,
				  LINES - (NUM_BANK + 4 + 1), 0)) == NULL) {
	    err(2, "newwin(streams)");
	}
	wborder(win_streams, ' ', ' ', '-', '-', '-', '-', '-', '-');
	mvwaddstr(win_streams, 1, 0, "Opened stream (`>': selected)");
	leaveok(win_streams, TRUE);
	touchwin(stdscr);
	wrefresh(win_streams);
	if ((win_status = subwin(stdscr, 1, COLS, LINES - 1, 0)) == NULL) {
	    err(2, "newwin(status)");
	}
	leaveok(win_status, TRUE);
	curs_set(0);
	touchwin(stdscr);
	wrefresh(win_status);
	cons = &curses_handler;
    } else {
	tcgetattr(fileno(fp_stdin), &t_orig);
	t_raw = t_orig;
	t_raw.c_lflag &= ~(ECHO|ICANON);
	tcsetattr(fileno(fp_stdin), TCSAFLUSH, &t_raw);
	cons = &dumb_handler;
    }

    atexit(clean_up);

    bzero(bank, sizeof(struct stream_bank) * NUM_BANK);
    num_opened = 0;

    for (cnt = 0; cnt < NUM_PGM; cnt++) {
	pgm[cnt].bank_no = EOF;
	pgm[cnt].start_frame = EOF;
	pgm[cnt].end_frame = EOF;
    }
    num_of_pgm = 0;
    cur_pgm = -1;

    if (listfile != NULL) {
	num_opened = load_list(listfile);
    }
    if (1 <= argc) {
	for (cnt = 0; cnt < NUM_BANK; cnt++) {
	    if (bank[cnt].fp == NULL) {
		bank[cnt].no = cnt;
		if (open_stream(&bank[cnt], *argv++) != NULL) {
/*
		    cur_bank = &bank[cnt];
*/
		    if (NUM_BANK <= ++num_opened || --argc < 1) {
			break;
		    }
		}
	    }
	}
    }
    bankno = 0;
    cur_bank = bank;
    fp = cur_bank->fp;
    frame_size = cur_bank->frame_size;
    num_frame = cur_bank->num_frame;
    cons->mark_stream_selected(cur_bank, 1);
    start_frame = 0L;
    end_frame = num_frame - 1;
    dir = 0;

    timeout_0.tv_sec = 0;
    timeout_0.tv_usec = 0;
    timeout_50.tv_sec = 0;
    timeout_50.tv_usec = 20000;
    timeout_60.tv_sec = 0;
    timeout_60.tv_usec = 16666;

    if (num_opened == 0) {
	cons->put_status("No DV stream file is opened");
    } else {
	cons->put_status("%d streams are opened", num_opened);
    }

    play_f = 0;
    repeat_f = 0;
    pgmmode_f = 0;
    skip = 0;
    div = 1;
    wait_frame = 0;
    eof_f = 0;
    exit_f = 0;

    num_fd = max(controller_fd, fileno(fp_stdin)) + 1;

    cons->put_action("Stop");

    do {
	cons->cursor_off();
#ifdef MSG_BY_SIGIO
	if (0 < msglen) {
	    cons->put_message(msgbuf, msglen);
	    msglen = 0;
/*
	    fcntl(msgpipefd[0], F_SETFL, O_NONBLOCK | O_ASYNC);
*/
	}
#endif
	FD_ZERO(&readfds);
	FD_SET(fileno(fp_stdin), &readfds);
	if (0 < controller_fd) {
	    FD_SET(controller_fd, &readfds);
	}
#ifndef MSG_BY_SIGIO
	FD_SET(msgpipefd[0], &readfds);
#endif
	select(num_fd, &readfds, NULL, NULL,
	       play_f ? (1 < div ? cur_bank->timeout : &timeout_0) : NULL);
	command = ACT_NONE;
#ifndef MSG_BY_SIGIO
	if (FD_ISSET(msgpipefd[0], &readfds)) {
	    fcntl(msgpipefd[0], F_SETFL, O_NONBLOCK);
	    while (0 < (msglen = read(msgpipefd[0], msgbuf, MSGBUFLEN))) {
		cons->put_message(msgbuf, msglen);
	    }
	    if (win_pgm != NULL) {
		touchwin(win_pgm);
		wrefresh(win_pgm);
	    }
	}
#endif /* MSG_BY_SIGIO */
	if (FD_ISSET(fileno(fp_stdin), &readfds)) {
	    command = get_key_event(fp_stdin);
	}
	if (command == ACT_NONE && 0 < controller_fd && FD_ISSET(controller_fd, &readfds)) {
	    command = controller_get_event(controller_fd);
	}
	if (command != ACT_NONE) {
	    cons->clear_status();
	    param = (char)(command & ACT_PARAM);

	    if (fp != NULL) {
		switch (command & ACT_MASK) {
		  case ACT_PAUSE:
		    if (play_f) {
			play_f = -1;
		    }
		    div = 1;
		    skip = 0;
		    cons->put_action("Pause");
		    break;
		  case ACT_PLAY:
		    if (param == 0) {
			if (play_f) {
			    if (div == 1 && skip == 0) {
				play_f = 0;
				cons->put_action("Pause");
				break;
			    }
			} else {
			    if (eof_f) {
				fseek(fp, 0L, SEEK_SET);
				eof_f = 0;
				frame_no = 0L;
			    }
			}
			param = 1;
		    }
		    cons->put_action_nonl("Play ");
		    play_f = 1;
		    if (0 < param) {
			skip = param - 1;
			div = 1;
			cons->put_action_term("X%d", param);
		    } else {
			skip = 0;
			div = -param;
			cons->put_action_term("1/%d", -param);
		    }
		    dir = 1;
		    break;
		  case ACT_BACK:
		    cons->put_action_nonl("Backward Play ");
		    play_f = 1;
		    if (0 < param) {
			skip = -(param + 1);
			div = 1;
			cons->put_action_term("X%d", param);
		    } else {
			skip = -2;
			div = -param;
			cons->put_action_term("1/%d", -param);
		    }
		    dir = -1;
		    break;
		  case ACT_STOP:
		    cons->put_action("Stop");
		    play_f = -1;
		    skip = 0;
		    fseek(fp, 0L, SEEK_SET);
		    break;
		  case ACT_NEXT:
		    cons->put_action("Next frame");
		    play_f = -1;
		    skip = 0;
		    dir = 1;
		    break;
		  case ACT_PREV:
		    cons->put_action("Previous frame");
		    play_f = -1;
		    skip = 0;
		    if (eof_f) {
			frame_no = num_frame - 1;
			fseeko(cur_bank->fp,
			       (off_t)frame_size * frame_no, SEEK_SET);
			eof_f = 0;
		    } else {
			if ((frame_no -= 2) < 0) {
			    frame_no = 0L;
			    fseek(cur_bank->fp, 0L, SEEK_SET);
			} else {
			    fseeko(cur_bank->fp, frame_size * -2, SEEK_CUR);
			}
		    }
		    dir = -1;
		    break;
		  case ACT_SKIP:
		    cons->put_action("Skip some frame");
		    cons->dialog(buf, FILENAME_MAX - 1,
				 "# of skip frame", "# of frame:");
		    frame_no += atol(buf);
		    if (num_frame <= frame_no) {
			frame_no = num_frame - 1;
		    } else if (frame_no < 0) {
			frame_no = 0;
		    }
		    fseeko(cur_bank->fp,
			   (off_t)frame_size * frame_no, SEEK_SET);
		    cons->put_status("Skip to frame %ld\n", frame_no);
		    break;
		  case ACT_JUMP:
		    cons->put_action("Jump to frame");
		    cons->dialog(buf, FILENAME_MAX - 1,
				 "Jump to frame #", "# of frame:");
		    frame_no = atol(buf);
		    if (num_frame <= frame_no) {
			frame_no = num_frame - 1;
		    } else if (frame_no < 0) {
			frame_no = 0;
		    }
/*
		    fseek(fp, 0L, SEEK_SET);
		    for (cnt = 0; cnt < frame_no; cnt++) {
			if (fseeko(fp, frame_size, SEEK_CUR) == EOF) {
			    cons->warn("frame pos = %ld", cnt);
			    break;
			}
			cons->put_status("frame = %ld\n", cnt + 1);
		    }
*/
		    fseeko(fp, (off_t)frame_size * frame_no, SEEK_SET);
		    cons->put_status("Jump to frame %ld\n", frame_no);
		    break;
		  case ACT_REPEAT:
		    if ((repeat_f = !repeat_f)) {
			cons->put_mode2("REP mode");
		    } else {
			cons->clear_mode2();
		    }
		    break;
		  case ACT_REFRESH:
		    cons->put_action("Refresh last frame");
		    play_f = -1;
		    skip = 0;
		    if (eof_f) {
			frame_no = num_frame - 1;
			fseeko(cur_bank->fp,
			       (off_t)frame_size * frame_no, SEEK_SET);
			eof_f = 0;
		    } else {
			if (--frame_no < 0) {
			    frame_no = 0L;
			    fseek(cur_bank->fp, 0L, SEEK_SET);
			} else {
			    fseeko(cur_bank->fp, frame_size * -1, SEEK_CUR);
			}
		    }
		    break;
		  case ACT_PGM:
		    switch (command) {
		      case ACT_PGM:
			if ((pgmmode_f = !pgmmode_f)) {
			    if (num_of_pgm <= 0) {
				pgmmode_f = 0;
			    } else {
				cons->put_mode1("PGM mode");
				if (win_pgm != NULL) {
				    cons->pgm_disp();
				}
			    }
			} else {
			    cons->clear_mode1();
			    cons->mark_stream_selected(cur_bank, 0);
			    cur_bank = &bank[bankno];
			    fp = cur_bank->fp;
			    frame_size = cur_bank->frame_size;
			    num_frame = cur_bank->num_frame;
			    if ((frame_no = cur_bank->last_pos) < 0) {
				frame_no = 0L;
			    }
			    cons->mark_stream_selected(cur_bank, 1);
			    if (fp != NULL) {
				fseeko(fp,
				       (off_t)frame_size * frame_no, SEEK_SET);
/*
				cons->put_status("DV stream is switched to #%d; "
						 "%s, %ld @ %ld frames\n",
						 bankno + 1, cur_bank->filename,
						 frame_no, num_frame);
*/
				eof_f = 0;
				play_f = -1;
				if (!pgmmode_f) {
				    cons->put_action(frame_no == 0 ? "Stop" : "Pause");
				}
			    }
			    start_frame = 0L;
			    end_frame = num_frame - 1;
			}
			break;
		      case ACT_PGM_IN:
			if (num_of_pgm <= 0) {
			    cur_pgm = 0;
			    num_of_pgm = 1;
			    pgm[cur_pgm].end_frame = EOF;
			} else {
			    if (pgm[cur_pgm].start_frame != EOF
				&& pgm[cur_pgm].end_frame != EOF
				&& (cur_pgm == num_of_pgm - 1
				    && num_of_pgm < NUM_PGM)) {
				cur_pgm = num_of_pgm++;
			    }
			    if (pgm[cur_pgm].bank_no != cur_bank->no) {
				pgm[cur_pgm].end_frame = EOF;
			    }
			}
			pgm[cur_pgm].bank_no = cur_bank->no;
			pgm[cur_pgm].start_frame = displayed_no;
			break;
		      case ACT_PGM_OUT:
			if (num_of_pgm <= 0) {
			    cur_pgm = 0;
			    num_of_pgm = 1;
			    pgm[cur_pgm].start_frame = EOF;
			} else {
			    if (pgm[cur_pgm].start_frame != EOF
				&& pgm[cur_pgm].end_frame != EOF
				&& (cur_pgm == num_of_pgm - 1
				    && num_of_pgm < NUM_PGM)) {
				cur_pgm = num_of_pgm++;
			    }
			    if (pgm[cur_pgm].bank_no != cur_bank->no) {
				pgm[cur_pgm].start_frame = EOF;
			    }
			}
			pgm[cur_pgm].bank_no = cur_bank->no;
			pgm[cur_pgm].end_frame = displayed_no;
			break;
		      case ACT_PGM_MOVE:
			cons->put_action("Move program");
			cons->dialog(buf, FILENAME_MAX - 1,
				     "Move program",
				     "Program %d to:", cur_pgm + 1);
			if ((tmp1 = atoi(buf) - 1) <= 0 || cur_pgm == tmp1) {
			    cons->put_action("Program moving canceled");
			    break;
			}
			tmp2 = pgm[cur_pgm].bank_no;
			ltmp1 = pgm[cur_pgm].start_frame;
			ltmp2 = pgm[cur_pgm].end_frame;
			if (cur_pgm < tmp1) {
			    for (cnt = cur_pgm + 1; cnt <= tmp1; cnt++) {
				pgm[cnt - 1].bank_no = pgm[cnt].bank_no;
				pgm[cnt - 1].start_frame = pgm[cnt].start_frame;
				pgm[cnt - 1].end_frame = pgm[cnt].end_frame;
			    }
			} else {
			    for (cnt = cur_pgm; tmp1 < cnt; cnt--) {
				pgm[cnt].bank_no = pgm[cnt - 1].bank_no;
				pgm[cnt].start_frame = pgm[cnt - 1].start_frame;
				pgm[cnt].end_frame = pgm[cnt - 1].end_frame;
			    }
			}
			pgm[tmp1].bank_no = tmp2;
			pgm[tmp1].start_frame = ltmp1;
			pgm[tmp1].end_frame = ltmp2;
			cons->put_status("Program %d moved to %d",
					 cur_pgm, tmp1);
			break;
		      case ACT_PGM_INS:
			if (NUM_PGM <= num_of_pgm || fp == NULL) {
			    break;
			}
			for (cnt = NUM_PGM - 1; cur_pgm < cnt; cnt--) {
			    pgm[cnt].bank_no = pgm[cnt - 1].bank_no;
			    pgm[cnt].start_frame = pgm[cnt - 1].start_frame;
			    pgm[cnt].end_frame = pgm[cnt - 1].end_frame;
			}
			pgm[cur_pgm].bank_no = cur_bank->no;
			pgm[cur_pgm].start_frame = EOF;
			pgm[cur_pgm].end_frame = EOF;
			cons->put_status("Program inserted to %d", cur_pgm);
			++num_of_pgm;
			break;
		      case ACT_PGM_DEL:
			for (cnt = cur_pgm + 1; cnt < NUM_PGM; cnt++) {
			    pgm[cnt - 1].bank_no = pgm[cnt].bank_no;
			    pgm[cnt - 1].start_frame = pgm[cnt].start_frame;
			    pgm[cnt - 1].end_frame = pgm[cnt].end_frame;
			}
			pgm[cnt - 1].bank_no = EOF;
			pgm[cnt - 1].start_frame = EOF;
			pgm[cnt - 1].end_frame = EOF;
			cons->put_status("Program %d deleted", cur_pgm);
			if (--num_of_pgm < 0) {
			    num_of_pgm = 0;
			}
			break;
		      case ACT_PGM_DISP:
			cons->pgm_disp();
			break;
		      case ACT_PGM_TOP:
			cur_pgm = 0;
			break;
		      case ACT_PGM_PREV:
			if (--cur_pgm < 0) {
			    cur_pgm = 0;
			}
			break;
		      case ACT_PGM_NEXT:
			++cur_pgm;
			break;
		      case ACT_PGM_LAST:
			cur_pgm = num_of_pgm - 1;
			break;
		    }
		    if (num_of_pgm <= cur_pgm) {
			cur_pgm = num_of_pgm - 1;
		    }
		    if (pgmmode_f
			|| (command != ACT_PGM && command != ACT_PGM_DISP)) {
			cons->pgm_refresh();
		    }
		    if (pgmmode_f && command != ACT_PGM_DISP) {
			cur_bank = &bank[pgm[cur_pgm].bank_no];
			start_frame = pgm[cur_pgm].start_frame;
			end_frame = pgm[cur_pgm].end_frame;
			frame_no = start_frame;
			fp = cur_bank->fp;
			frame_size = cur_bank->frame_size;
			num_frame = cur_bank->num_frame;
			if (fp != NULL) {
			    fseeko(fp, (off_t)frame_size * frame_no, SEEK_SET);
/*
			    cons->put_status("DV stream is switched to #%d; "
					     "%s, %ld @ %ld frames\n",
					     bankno + 1, cur_bank->filename,
					     frame_no, num_frame);
*/
			    eof_f = 0;
			    play_f = -1;
			    if (!pgmmode_f) {
				cons->put_action(frame_no == 0 ? "Stop" : "Pause");
			    }
			}
		    }
		    break;
		  case ACT_WRITE:
		    cons->put_action("Write stream to file");
		    cons->dialog(buf, FILENAME_MAX - 1,
				 "Write file", "Output file:");
		    if ((fo = fopen(buf, "w")) == NULL) {
			cons->warn("Write file: %s", buf);
			break;
		    } else {
			long frame_pos, start_frame, end_frame, frame_size;
			long num_frame = 0L;

			for (cnt = 0; cnt < num_of_pgm; cnt++) {
			    if (pgm[cnt].bank_no == EOF) {
				continue;
			    }
			    fi = bank[pgm[cnt].bank_no].fp;
			    if ((start_frame = pgm[cnt].start_frame) == EOF) {
				start_frame = 0L;
			    }
			    if ((end_frame = pgm[cnt].end_frame) == EOF) {
				end_frame = bank[pgm[cnt].bank_no].num_frame - 1;
			    }
			    frame_size = bank[pgm[cnt].bank_no].frame_size;
			    fseeko(fi, (off_t)frame_size * start_frame,
				   SEEK_SET);
			    for (frame_pos = start_frame;
				 frame_pos <= end_frame; frame_pos++) {
				cons->put_status("Pgm #%d, frame %ld",
						 cnt + 1, frame_pos);
				fread(frame, frame_size, 1, fi);
				fwrite(frame, frame_size, 1, fo);
				++num_frame;
			    }
			}
			fclose(fo);
			cons->put_status("Wrote to %s, %ld frames",
					 buf, num_frame);
		    }

		    frame_no = 0L;
		    play_f = 0;
		    cons->put_action("Stop");
		    break;
		  case ACT_OUT:
		    cons->put_action("Output stream to i.LINK");
		    pipe(outpipe);
		    if ((fwcontrol_pid = fork()) == 0) {
			/* child */
			char *cmd = strdup(fwcontrol_cmd), *ptr, **arg;
			int argc, cnt;

			if (cmd == NULL) {
			    cons->warn("fork(1)");
			    return 1;
			}
			for (argc = 1, ptr = cmd;
			     (ptr = strchr(ptr, ' ')) != NULL; argc++, ptr++) {
			    ;
			}
#ifdef DEBUG
			fprintf(stderr, "argc = %d\n", argc);
#endif
			if ((arg = malloc(sizeof(char *) * (argc + 1))) == NULL) {
			    cons->warn("fork(2)");
			    return 1;
			}
			for (cnt = 0, ptr = cmd; cnt < argc; cnt++) {
			    arg[cnt] = ptr;
			    if ((ptr = strchr(ptr, ' ')) != NULL) {
				*ptr++ = '\0';
			    }
#ifdef DEBUG
			    fprintf(stderr, "arg[%d] = `%s'\n", cnt, arg[cnt]);
#endif
			}
			arg[argc++] = NULL;
			if ((ptr = strrchr(arg[0], '/')) != NULL) {
			    arg[0] = ++ptr;
			}
#ifdef DEBUG
			fprintf(stderr, "cmd = `%s'\n", cmd);
			for (cnt = 0, ptr = cmd; cnt < argc; cnt++) {
			    fprintf(stderr, "arg[%d] = `%s'\n", cnt, arg[cnt]);
			}
#endif
			dup2(outpipe[0], 0);
			close(outpipe[0]);
			dup2(msgpipefd[1], 1);
			dup2(msgpipefd[1], 2);
			close(msgpipefd[1]);
			execv(cmd, arg);
			cons->warn("execv");
			return 2;
		    }

		    /* parent */
		    close(outpipe[0]);
		    do {
			long frame_pos, start_frame, end_frame, frame_size;
			long num_frame = 0L;
			int tmp;

			for (cnt = 0; cnt < num_of_pgm; cnt++) {
			    if (pgm[cnt].bank_no == EOF) {
				continue;
			    }
			    fi = bank[pgm[cnt].bank_no].fp;
			    if ((start_frame = pgm[cnt].start_frame) == EOF) {
				start_frame = 0L;
			    }
			    if ((end_frame = pgm[cnt].end_frame) == EOF) {
				end_frame = bank[pgm[cnt].bank_no].num_frame - 1;
			    }
			    frame_size = bank[pgm[cnt].bank_no].frame_size;
			    fseeko(fi, (off_t)frame_size * start_frame,
				   SEEK_SET);
			    for (frame_pos = start_frame;
				 frame_pos <= end_frame; frame_pos++) {
				cons->put_status("Pgm #%d, frame %ld",
						 cnt + 1, frame_pos);
				fread(frame, frame_size, 1, fi);
				if (write(outpipe[1], frame, frame_size) < frame_size) {
				    kill(fwcontrol_pid, SIGTERM);
				    waitpid(fwcontrol_pid, &tmp, WNOHANG);
				    break;
				}				    
				++num_frame;
			    }
			}
			close(outpipe[1]);
			cons->put_status("Wrote to %s, %ld frames",
					 buf, num_frame);
		    } while (repeat_f);

		    frame_no = 0L;
		    play_f = 0;
		    cons->put_action("Stop");
		    break;
		}
	    }
	    switch (command & ACT_MASK) {
	      case ACT_LOAD:
		cons->filename_dialog(buf, FILENAME_MAX - 1,
				      "Open list", "Stream/Program file:");
		num_opened = load_list(buf);
		bankno = cur_bank->no;
		if (pgmmode_f) {
		    cur_bank = &bank[pgm[cur_pgm].bank_no];
		}
		fp = cur_bank->fp;
		frame_size = cur_bank->frame_size;
		num_frame = cur_bank->num_frame;
		frame_no = cur_bank->last_pos = 0L;
		if (pgmmode_f) {
		    frame_no = pgm[cur_pgm].start_frame;
		}
		cons->mark_stream_selected(cur_bank, 1);
		fseeko(fp, (off_t)frame_size * frame_no, SEEK_SET);
		start_frame = 0L;
		end_frame = num_frame - 1;
		cons->put_status("DV stream is switched to #%d; "
				 "%s, %ld @ %ld frames\n",
				 bankno + 1, cur_bank->filename,
				 frame_no, num_frame);
		eof_f = 0;
		break;
	      case ACT_SAVE:
		cons->filename_dialog(buf, FILENAME_MAX - 1,
				      "Save list", "Stream/Program file:");
		save_list(buf);
		break;
	      case ACT_OPEN:
		play_f = 0;
		cons->put_action(frame_no == 0 ? "Stop" : "Pause");
		cons->clear_frame_info();
		if (param < 0) {
		    if (command == ACT_NFILE) {
			param = bankno + 2;
		    } else if (command == ACT_PFILE) {
			if ((param = bankno) < 1) {
			    param = 1;
			}
		    } else {
			param = 0;
		    }
		}
		if (0 < param && param <= NUM_BANK) {
		    cons->mark_stream_selected(&bank[bankno], 0);
		    if (eof_f) {
			bank[bankno].last_pos = 0L;
		    } else {
			bank[bankno].last_pos = frame_no - 1;
		    }
		    bankno = param - 1;
		    cur_bank = &bank[bankno];
		    cur_bank->no = bankno;
		    fp = cur_bank->fp;
		    frame_size = cur_bank->frame_size;
		    num_frame = cur_bank->num_frame;
		    if ((frame_no = cur_bank->last_pos) < 0) {
			frame_no = 0L;
		    }
		    cons->mark_stream_selected(cur_bank, 1);
		    if (fp != NULL) {
			fseeko(fp, (off_t)frame_size * frame_no, SEEK_SET);
			cons->put_status("DV stream is switched to #%d; "
				 "%s, %ld @ %ld frames\n",
			       bankno + 1, cur_bank->filename,
			       frame_no, num_frame);
			eof_f = 0;
			cons->put_action(frame_no == 0 ? "Stop" : "Pause");
			break;
		    }
		} else {
		    if (cur_bank->fp != NULL) {
			fclose(cur_bank->fp);
			fp = cur_bank->fp = NULL;
			if (cur_bank->filename != NULL) {
			    free(cur_bank->filename);
			    cur_bank->filename = NULL;
			}
			--num_opened;
		    }
		}
		cons->show_stream_info(&bank[bankno]);
		cons->filename_dialog(buf, FILENAME_MAX - 1,
				      "Open file",
				      "DV stream file #%d:", bankno + 1);
		bank[bankno].no = bankno;
		cur_bank = &bank[bankno];
		if (open_stream(cur_bank, buf) != NULL) {
		    ++num_opened;
		    fp = cur_bank->fp;
		    frame_size = cur_bank->frame_size;
		    num_frame = cur_bank->num_frame;
		    frame_no = 0L;
		}
		play_f = 0;
		cons->put_action(frame_no == 0 ? "Stop" : "Pause");
		skip = 0;
		start_frame = 0L;
		end_frame = num_frame - 1;
		break;
	      case ACT_DISP:
		if ((verbose_f = !verbose_f)) {
		    cons->put_status("Verbose: On");
		} else {
		    cons->put_status("Verbose: Off");
		    cons->clear_frame_info();
		}
		break;
	      case ACT_QUIT:
		play_f = 0;
		exit_f = 1;
		break;
	    }
	}

	if (play_f && fp == NULL) {
	    cons->put_status("No DV stream file is assumed");
	    play_f = 0;
	}

	if (play_f) {
	    if (wait_frame++ < div) {
		continue;
	    }
	    wait_frame = 0;
	    if (fp == NULL) {
		cons->put_status("Not opened");
		++eof_f;
		play_f = -1;
		cons->put_action("Stop");
	    } else if (fread(frame, frame_size, 1, fp) < 1) {
		if (repeat_f) {
		    frame_no = 0L;
		    fseek(fp, 0L, SEEK_SET);
		} else {
		    cons->put_status("EOF      ");
		    ++eof_f;
		    play_f = -1;
		    cons->put_action("Stop");
		}
	    } else {
		cons->put_frame_no(frame_no);
		if (verbose_f) {
		    parse_subcode(frame, cur_bank->system_is_625);
		}
		cons->finish_frame_info();
		if (!nodisp_f) {
		    if (write(pipefd[1], frame, frame_size) < frame_size) {
			++exit_f;
		    }
		}
		displayed_no = frame_no;
		if ((frame_no += (skip + 1)) < 0 && !pgmmode_f) {
		    frame_no = 0L;
		    fseek(fp, 0L, SEEK_SET);
		    play_f = 0;
		    cons->put_action("Pause");
		} else if (skip != 0) {
		    fseeko(fp, (off_t)frame_size * skip, SEEK_CUR);
		}
		if (pgmmode_f) {
		    if (frame_no < start_frame || end_frame < frame_no
			|| (command == ACT_PREV && displayed_no <= start_frame)
			|| (command == ACT_NEXT && end_frame <= displayed_no)) {
			if (dir < 0) {
#if 0
			    while (0 <= --cur_pgm
				   && (pgm[cur_pgm].start_frame == EOF
				       || pgm[cur_pgm].end_frame == EOF)) {
				cons->put_status("Ignore incomplete program %d\n", cur_pgm);
			    }
#else
			    --cur_pgm;
#endif
			    if (cur_pgm < 0) {
				cur_pgm = 0;
				eof_f = 1;
				play_f = 0;
			    } else {
				cur_bank = &bank[pgm[cur_pgm].bank_no];
				frame_no = -1;
			    }
			} else {
#if 0
			    while (++cur_pgm < num_of_pgm
				   && (pgm[cur_pgm].start_frame == EOF
				       || pgm[cur_pgm].end_frame == EOF)) {
				cons->put_status("Ignore incomplete program %d\n", cur_pgm);
			    }
#else
			    ++cur_pgm;
#endif
			    if (num_of_pgm <= cur_pgm) {
				cur_pgm = num_of_pgm - 1;
				eof_f = 1;
				play_f = 0;
			    } else {
				cur_bank = &bank[pgm[cur_pgm].bank_no];
				frame_no = 1;
			    }
			}
			fp = cur_bank->fp;
			frame_size = cur_bank->frame_size;
			num_frame = cur_bank->num_frame;
			if ((start_frame = pgm[cur_pgm].start_frame) == EOF) {
			    start_frame = 0L;
			}
			if ((end_frame = pgm[cur_pgm].end_frame) == EOF) {
			    end_frame = num_frame - 1;
			}
			if (frame_no < 0) {
			    frame_no = end_frame;
			} else {
			    frame_no = start_frame;
			}
			if (!eof_f && fp != NULL) {
			    fseeko(fp, (off_t)frame_size * frame_no, SEEK_SET);
			} else {
			    cons->put_action("Stop");
			}
			cons->put_status("Current Pgm # = %d", cur_pgm + 1);
			cons->pgm_refresh();
		    }
		}
	    }
	    if (play_f < 0) {
		++play_f;
	    }
	}
    } while (!exit_f);

    return 0;
}


FILE
*open_stream(struct stream_bank *bank, const char *filename)
{
    u_char frame[DIF_BLK_SIZE];
    struct stat stat;

    if (bank->filename != NULL) {
	free(bank->filename);
    }
    bank->filename = strdup(filename);
    if ((bank->fp = fopen(bank->filename, "r")) == NULL) {
	cons->warn(bank->filename);
    } else {
	fread(frame, DIF_BLK_SIZE, 1, bank->fp);
	if ((frame[0] & 0xe0) != 0) {
	    /* Header is not found */
	    cons->put_status("%s: broken DV stream", bank->filename);
	    fclose(bank->fp);
	    bank->fp = NULL;
	} else {
	    fseek(bank->fp, 0L, SEEK_SET);
	    if ((bank->system_is_625 = (frame[3] & 0x80)) == 0) {
		bank->timeout = &timeout_60;
		bank->frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_525);
	    } else {
		bank->timeout = &timeout_50;
		bank->frame_size = (DIF_SEQ_SIZE * NUM_DIF_SEQ_625);
	    }
	    fstat(fileno(bank->fp), &stat);
	    bank->num_frame = stat.st_size / bank->frame_size;
	    bank->last_pos = 0L;
	    cons->show_stream_info(bank);
	}
    }

    if (bank->fp == NULL && bank->filename != NULL) {
	free(bank->filename);
	bank->filename = NULL;
    }

    return bank->fp;
}


int
load_list(const char *filename)
{
    int no, cnt, value, num_opened = 0;
    long l_value, num_frame;
    int info;
    FILE *fp;
    char buf[1024], *ptr, *next;

    if ((fp = fopen(filename, "r")) == NULL) {
	cons->warn(filename);
	return num_opened;
    }

    for (cnt = 0; cnt < NUM_BANK; cnt++) {
	if (bank[cnt].fp != NULL) {
	    fclose(bank[cnt].fp);
	}
	bank[cnt].fp = NULL;
	if (bank[cnt].filename != NULL) {
	    free(bank[cnt].filename);
	}
	bank[cnt].filename = NULL;
	bank[cnt].system_is_625 = 0;
	bank[cnt].frame_size = 0;
	bank[cnt].num_frame = 0;
	bank[cnt].last_pos = 0L;
	bank[cnt].timeout = NULL;
    }

    info = 0;
    for (cnt = 1; fgets(buf, 1024, fp) != NULL; cnt++) {
	if ((ptr = strchr(buf, '#')) != NULL) {
	    *ptr = '\0';
	}
	if ((ptr = strchr(buf, '\n')) != NULL) {
	    *ptr = '\0';
	}
	if (!strncmp(buf, "Streams:", 8)) {
	    info = 1;
	    continue;
	}
	if ((ptr = strchr(buf, ':')) == NULL) {
	    continue;
	}
	*ptr++ = '\0';
	switch (info) {
	  case 1:
	    if (!strncmp(buf, "Programs", 8)) {
		info = 2;
		continue;
	    }
	    if ((no = atoi(buf)) < 0 || NUM_BANK <= no) {
		cons->put_status("Ignore illegal bank number (%d)"
				 " at line (%d)\n", no, cnt);
	    } else {
		bank[no].no = no;
		if (open_stream(&bank[no], ptr) != NULL) {
		    ++num_opened;
		}
	    }
	    break;
	  case 2:
	    if ((no = atoi(buf)) < 0 || NUM_PGM <= no) {
		cons->put_status("Ignore illegal program number (%d)"
				 " at line (%d)\n", no, cnt);
		continue;
	    }
	    pgm[no].bank_no = EOF;
	    pgm[no].start_frame = EOF;
	    pgm[no].end_frame = EOF;
	    if ((next = strchr(ptr, ',')) != NULL) {
		*next++ = '\0';
	    } else {
		cons->put_status("Ignore illegal line"
				 " (both IN nor OUT frame are not"
				 " specified) at line (%d)\n", cnt);
		continue;
	    }
	    if ((value = atoi(ptr)) < 0 || NUM_BANK <= value) {
		cons->put_status("Ignore illegal bank number"
				 " (%d) at line (%d)\n", value, cnt);
		continue;
	    }
	    pgm[no].bank_no = value;
	    num_frame = bank[pgm[no].bank_no].num_frame;
	    ptr = next;
	    if ((next = strchr(ptr, ',')) != NULL) {
		*next++ = '\0';
	    }
	    if ((l_value = atol(ptr)) != EOF
		&& (l_value < 0 || num_frame <= l_value)) {
		cons->put_status("Ignore illegal start frame"
				 " (%ld) at line (%d)\n", l_value, cnt);
		pgm[no].start_frame = EOF;
	    } else {
		pgm[no].start_frame = l_value;
	    }
	    if (next != NULL) {
		if ((l_value = atol(next)) != EOF
		    && (l_value < 0 || num_frame <= l_value)) {
		    cons->put_status("Ignore illegal end frame"
				     " (%ld) at line (%d)\n", l_value, cnt);
		    pgm[no].end_frame = EOF;
		} else {
		    pgm[no].end_frame = l_value;
		}
	    }
/*
	    printf("%d: %d, %ld/%ld\n", no, pgm[no].bank_no,
		   pgm[no].start_frame, pgm[no].end_frame);
*/
	    break;
	  default:
	    /* do nothing */
	    break;
	}
    }

    fclose(fp);

    for (cnt = 0; cnt < NUM_BANK; cnt++) {
	if (bank[cnt].fp != NULL) {
	    cur_bank = &bank[cnt];
	    break;
	}
    }

    cur_pgm = EOF;
    for (cnt = 0; cnt < NUM_PGM; cnt++) {
	if (pgm[cnt].bank_no != EOF) {
	    if (cur_pgm == EOF) {
		cur_pgm = cnt;
	    }
	    num_of_pgm = cnt + 1;
	}
    }

    cons->put_status("Loading %s done", filename);

    return num_opened;
}


void
save_list(const char *filename)
{
    int cnt;
    FILE *fp;

    if ((fp = fopen(filename, "w")) == NULL) {
	cons->warn(filename);
	return;
    }

    fputs("Streams:\n", fp);
    for (cnt = 0; cnt < NUM_BANK; cnt++) {
	if (bank[cnt].fp != NULL) {
	    fprintf(fp, "%d:%s\n", bank[cnt].no, bank[cnt].filename);
	}
    }

    fputs("\nPrograms:\n", fp);
    for (cnt = 0; cnt < NUM_PGM; cnt++) {
	if (pgm[cnt].bank_no != EOF) {
	    fprintf(fp, "%d: %d, %ld, %ld\n", cnt, pgm[cnt].bank_no,
		    pgm[cnt].start_frame, pgm[cnt].end_frame);
	}
    }

    fclose(fp);

    cons->put_status("Wrote %s", filename);
}


struct key_table {
    int chr;
    int action;
};

struct key_table key_table_playmode[] = {
    { 'p', ACT_PLAY },
    { '8', ACT_PLAYx8 },
    { '4', ACT_PLAYx4 },
    { '2', ACT_PLAYx2 },
    { '1', ACT_PLAYx1 },
    { '9', ACT_PLAYs5 },
    { '0', ACT_PLAYs10 },
    { ')', ACT_BACKs10 },
    { '(', ACT_BACKs5 },
    { '!', ACT_BACKx1 },
    { '@', ACT_BACKx2 },
    { '$', ACT_BACKx4 },
    { '*', ACT_BACKx8 },
    { 's', ACT_STOP },
    { 'n', ACT_NEXT },
    { 'b', ACT_PREV },
    { 'j', ACT_JUMP },
    { 'J', ACT_SKIP },
    { 'r', ACT_REFRESH },
    { 'R', ACT_REPEAT },
    { ' ', ACT_PLAY },
    { '\r', ACT_NEXT },
    { KEY_BACKSPACE, ACT_PREV },
    { 'o', ACT_OPEN },
    { 'A', ACT_OPEN_1 },
    { 'B', ACT_OPEN_2 },
    { 'C', ACT_OPEN_3 },
    { 'D', ACT_OPEN_4 },
    { 'E', ACT_OPEN_5 },
    { 'F', ACT_OPEN_6 },
    { 'G', ACT_OPEN_7 },
    { 'H', ACT_OPEN_8 },
    { KEY_F( 1), ACT_OPEN_1 },
    { KEY_F( 2), ACT_OPEN_2 },
    { KEY_F( 3), ACT_OPEN_3 },
    { KEY_F( 4), ACT_OPEN_4 },
    { KEY_F( 5), ACT_OPEN_5 },
    { KEY_F( 6), ACT_OPEN_6 },
    { KEY_F( 7), ACT_OPEN_7 },
    { KEY_F( 8), ACT_OPEN_8 },
    { KEY_NPAGE, ACT_NFILE },
    { KEY_PPAGE, ACT_PFILE },
    { 'P',	 ACT_PGM },
    { 'V',	 ACT_PGM_DISP },
    { 'I',	 ACT_PGM_IN },
    { 'O',	 ACT_PGM_OUT },
    { 'M',	 ACT_PGM_MOVE },
    { 'i',	 ACT_PGM_INS },
    { 'd',	 ACT_PGM_DEL },
    { KEY_F( 9), ACT_PGM },
    { KEY_F(10), ACT_PGM_DISP },
    { KEY_F(11), ACT_PGM_IN },
    { KEY_F(12), ACT_PGM_OUT },
    { KEY_IC,	 ACT_PGM_INS },
    { KEY_DC,	 ACT_PGM_DEL },
    { 'T',	 ACT_PGM_TOP },
    { 'U',	 ACT_PGM_PREV },
    { 'N',	 ACT_PGM_NEXT },
    { 't',	 ACT_PGM_LAST },
    { KEY_LEFT,	 ACT_PGM_TOP },
    { KEY_UP,	 ACT_PGM_PREV },
    { KEY_DOWN,	 ACT_PGM_NEXT },
    { KEY_RIGHT, ACT_PGM_LAST },
    { 'v', ACT_DISP },
    { 'L', ACT_LOAD },
    { 'S', ACT_SAVE },
    { 'w', ACT_WRITE },
    { 'W', ACT_OUT },
    { 'q', ACT_QUIT },
    { 'x', ACT_QUIT },
};

struct key_table *key_table = key_table_playmode;


int
get_key_event(FILE *fi)
{
    int c;
    struct key_table *ptr;

    c = cons->get_char(fi);
/*
    cons->put_status("key=%o\n", c);
*/
    for (ptr = key_table; 0 < ptr->chr; ptr++) {
	if (c == ptr->chr) {
	    return ptr->action;
	}
    }
    return ACT_NONE;
}


void
parse_subcode(u_char *frame, int system_is_625)
{
    int seq, blk, cnt, tmp;
    int num_seq = system_is_625 ? NUM_DIF_SEQ_625 : NUM_DIF_SEQ_525;
    int timecode_f = 0, datacode_f = 0, src_f = 0, date_f = 0, time_f = 0;
    u_char *DIF = frame, *ptr, *PC;
    char *source = NULL, buf[10], date[25], time[15];

    for (seq = 0; seq < num_seq; seq++) {
	for (blk = 0; blk < DIF_NUM_BLKS_IN_SEQ; blk++) {
	    switch ((DIF[0] >> 5) & 0x7) {
	      case 0:	/* Header */
		break;
	      case 1:	/* Subcode */
		if (!timecode_f) {
		    for (ptr = &DIF[3], cnt = 0; cnt < 6; cnt++, ptr += 8) {
			PC = &ptr[3];
			if (PC[0] == 0x13) {
			    /* Timecode pack */
			    cons->disp_timecode(((PC[4] >> 4) & 0x03) * 10
						+ (PC[4] & 0x0f),
						((PC[3] >> 4) & 0x07) * 10
						+ (PC[3] & 0x0f),
						((PC[2] >> 4) & 0x07) * 10
						+ (PC[2] & 0x0f),
						((PC[1] >> 4) & 0x03) * 10
						+ (PC[1] & 0x0f),
						system_is_625,
						(PC[1] & 0x40));
			    ++timecode_f;
			    break;
			}
		    }
		}
		break;
	      case 2:	/* VAUX */
		if (!datacode_f) {
		    for (ptr = &DIF[3], cnt = 0; cnt < 15; cnt++, ptr += 5) {
			PC = ptr;
			if (PC[0] == 0x60) {
			    /* VAUX Source pack */
			    if (!src_f) {
				tmp = PC[2] & 0x07;
				if (PC[1] == 0xff && tmp == 0x07) {
				    source = "---";
				} else {
				    tmp = tmp * 100
					+ (PC[1] >> 4) * 10
					+ (PC[1] & 0x0f);
				    switch (PC[3] >> 6) {
				      case 0:	/* Camera */
					source = "Camera";
					break;
				      case 1:	/* Line */
					sprintf(buf, "LINE %d", tmp);
					source = buf;
					break;
				      case 2:	/* CATV */
					sprintf(buf, "CH C%d", tmp);
					source = buf;
					break;
				      case 3:	/* TV */
					sprintf(buf, "CH %d", tmp);
					source = buf;
					break;
				    }
				}
				++src_f;
			    }
			} else if (PC[0] == 0x62) {
			    /* VAUX Rec date pack */
			    if (!date_f) {
				sprintf(date, "%x%x/%02x/%d%x",
				       (PC[4] >> 4) & 0x0f,
					PC[4] & 0x0f,
					PC[3] & 0x0f,
				       (PC[2] >> 4) & 0x03,
					PC[2] & 0x0f);
				++date_f;
			    }
			} else if (PC[0] == 0x63) {
			    /* VAUX Rec time pack */
			    if (!time_f) {
				if ((PC[1] & 0x3f) == 0x3f) {
				    tmp = 0;
				} else {
				    tmp = ((PC[1] >> 4) & 0x03) * 10
					+ (PC[1] & 0x0f);
				}
				sprintf(time, "%d%x:%d%x:%d%x.%02d",
					(PC[4] >> 4) & 0x03,
					 PC[4] & 0x0f,
					(PC[3] >> 4) & 0x07,
					 PC[3] & 0x0f,
					(PC[2] >> 4) & 0x07,
					 PC[2] & 0x0f,
					 tmp);
				++time_f;
				break;
			    }
			}
			if (src_f && date_f && time_f) {
			    cons->disp_datacode(date, time, source);
			    ++datacode_f;
			    break;
			}
		    }
		}
		break;
	      case 3:	/* Audio */
		break;
	      case 4:	/* Video */
		break;
	    }
	    DIF += DIF_BLK_SIZE;
	}
    }
}


void
usage(char *argv0)
{
    fprintf(stderr, "Usage:\t%s [options] [stream file]\n", argv0);
    fprintf(stderr, "\toptions:\n");
    fprintf(stderr, "\t   -c [<controller path>][:<controller type>]\n");
    fprintf(stderr, "\t\tUSB HID controller $B$N(B path $B$d(B type $B$r;XDj$9$k(B\n");
    fprintf(stderr, "\t\tPath $B$H(B type $B$O$I$A$i$+$,>JN,2DG=(B\n");
    fprintf(stderr, "\t\tDefault $B$O$=$l$>$l(B `%s' $B$H(B `%s'\n",
					UHID_DEVNAME, UHID_DEVTYPE);
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -f <cmd>\n");
    fprintf(stderr, "\t\tFwcontrol program $B$N(B command line $B$r;XDj$9$k(B\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -l <list file>\n");
    fprintf(stderr, "\t\t$B5/F0;~$K(B bank / program list file $B$rFI$_9~$^$;$k(B\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -n\tPlaydv $B$K$h$k2hLL=PNO$rM^;_$9$k(B(debug $BMQ!K(B\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -p <path>\n");
    fprintf(stderr, "\t\tPlaydv program $B$N(B path $B$r;XDj$9$k(B\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "\t   -v\tStream $B$N:F@8Cf$K(B timecode $B$d(B datacode $B$rI=<($9$k(B\n");
    fprintf(stderr, "\t   -w\tConsole $B$G$NI=<($K(B curses $B$r;HMQ$9$k(B\n");

    exit(1);
}
